SuperStrict


'-----Imports-----
Framework BRL.GLMax2D
Import BRL.Random
Import BRL.StandardIO
Include "Octree.bmx"
Include "Simplex.bmx"



Const w:Int = 900	'Fensterbreite
Const h:Int = 600	'Fensterhöhe
Const p:Int = 6 	'Pixelgröße

wld.Init(8)
cam.Init(p)


'-----Hauptschleife-----
Repeat 
	cam.Control()
	cam.Render()
	cam.Draw(0, 0, p)

	'Infotext
	DrawText "+/-: Scrollrad", 15, 15
	DrawText "SPEED: "+cam.spd, 15, 30
	
	DrawText "Einstellungen +/-: F1-F3, Space + F1-F3", 15, 60
	DrawText "Sichtweite : "+cam.vis, 15, 75
	DrawText "Sichtfeld  : "+cam.fov, 15, 90
	DrawText "Detaillevel: "+wld.lod, 15, 105
	
	DrawText "Octree: "+wld.size+"^3", 15, 150
	DrawText "Voxels: "+wld.voxcnt   , 15, 165
	DrawText "Memory: "+(GCMemAlloced()/1000000)+" MB", 15, 180
	DrawText "fps: "+GetFps(), 15, 210
	
	'Screen Update
	Flip(0)
	Cls()
Until KeyHit(KEY_ESCAPE) Or AppTerminate()

EndGraphics()
End



'-----Types und Funktionen-----
Type cam
	Global pix:TPixmap		'Pixmap
	Global asp:Float		'Seitenverhältnis
	
	Global pos:Float[3]		'Ortsvektor
	Global dir:Float[3]		'Richtungsvektor
	Global upv:Float[3]		'Up-Vektor
	Global crs:Float[3]		'Kreuzprodukt
	Global ang:Float[2]		'Drehwinkel
	Global spd:Int 			'Geschwindigkeit
	
	Global vis:Int			'Sichtweite
	Global fov:Int 			'Field of View
	Global sfov:Float		'Sinus fov
	
	
	'Initialisierung
	Function Init(r:Int)
		'Fenster öffnen
		SetGraphicsDriver(GLMax2DDriver())
		Graphics(w, h, 0, 60)
		
		'Maus unsichtbar
		HideMouse()
		
		'Pixmap erstellen
		pix = CreatePixmap(w/r, h/r, PF_RGBA8888)
		ClearPixels(pix, 0)
		
		asp = Float(pix.height)/pix.width
		
		'Position auf die Mitte der Umgebung setzen
		pos[0] = wld.size/2
		pos[1] = wld.size/2
		pos[2] = wld.size/2
		
		'Einstellungen
		vis = 50
		fov = 45.0
		sfov = Sin(fov)
		spd = 10
	EndFunction
	
	
	'Kamera bewegen
	Function Control()
		Global delta:Int
		Global oms:Int
		
		Local mxs:Float
		Local mys:Float
		Local mzs:Int
		
		Local speed:Float
		
		delta = MilliSecs()-oms
		oms = MilliSecs()
		
		'Speed mit Mausrad verändern
		mzs = MouseZSpeed()
		If mzs > 0 Then
			spd :+ mzs^4 
		Else
			spd :- mzs^4
		EndIf
		If spd < 0 Then spd = 0
		
		'Blickrichtung mit Maus ändern
		mxs = MouseX()-w/2
		mys = MouseY()-h/2
		MoveMouse w/2, h/2
		ang[0] :+ mys/10
		ang[1] :+ mxs/10
		Rotate(ang)
		
		'Kreuzprodukt des Richtungs- und Up-Vektors		
		crs[0] = dir[1]*upv[2] - dir[2]*upv[1]
		crs[1] = dir[2]*upv[0] - dir[0]*upv[2]
		crs[2] = dir[0]*upv[1] - dir[1]*upv[0]
		
		'Kamera bewegen
		speed = Float(spd*delta)/1000
		
		If KeyDown(KEY_W) Then Move(dir, speed)
		If KeyDown(KEY_S) Then Move(dir,-speed)
		If KeyDown(KEY_A) Then Move(crs, speed)
		If KeyDown(KEY_D) Then Move(crs,-speed)
		If KeyDown(KEY_R) Then Move(upv, speed)
		If KeyDown(KEY_F) Then Move(upv,-speed)
		
		'Kamera einstellen
		If KeyDown(KEY_SPACE) Then
			If KeyDown(KEY_F1) Then vis :+ 1*delta
			If KeyDown(KEY_F2) Then fov :+ 1; sfov = Sin(fov)
			If KeyHit(KEY_F3)  Then wld.lod :+ 1
		Else 
			If KeyDown(KEY_F1) Then vis :- 1*delta
			If KeyDown(KEY_F2) Then fov :- 1; sfov = Sin(fov)
			If KeyHit(KEY_F3)  Then wld.lod :- 1
		EndIf
		
		If wld.lod > wld.maxlod Then wld.lod = wld.maxlod
		If wld.lod < 1          Then wld.lod = 1
		
		'Kamera auf Größe der Umgebung begrenzen
		If pos[0] < 0.01          Then pos[0] = 0.01
		If pos[1] < 0.01          Then pos[1] = 0.01
		If pos[2] < 0.01          Then pos[2] = 0.01
		If pos[0] > wld.size-0.01 Then pos[0] = wld.size-0.01
		If pos[1] > wld.size-0.01 Then pos[1] = wld.size-0.01
		If pos[2] > wld.size-0.01 Then pos[2] = wld.size-0.01
	EndFunction
	
	
	'Bild Rendern
	Function Render()
		Local pixptr:Byte Ptr
		Local raydir:Float[3]
		Local vstep:Float[3]
		Local hstep:Float[3]
		Local r:Byte, g:Byte, b:Byte
		
		'Schrittgröße in horizontaler Richtung
		hstep[0] = crs[0]*sfov*2/(pix.width-1)
		hstep[1] = crs[1]*sfov*2/(pix.width-1)
		hstep[2] = crs[2]*sfov*2/(pix.width-1)
		
		'Schrittgröße in vertikaler Richtung
		vstep[0] = upv[0]*sfov*asp*2/(pix.height-1)
		vstep[1] = upv[1]*sfov*asp*2/(pix.height-1)
		vstep[2] = upv[2]*sfov*asp*2/(pix.height-1)
		
		'Richtung des Strahls
		raydir[0] = dir[0]+crs[0]*sfov+upv[0]*sfov*asp
		raydir[1] = dir[1]+crs[1]*sfov+upv[1]*sfov*asp
		raydir[2] = dir[2]+crs[2]*sfov+upv[2]*sfov*asp
		
		'Pointer zur Pixmap
		pixptr = pix.Pixels
		
		'Jeden Pixel Raytracen
		For Local y:Int = 0 Until pix.height
			If y > 0 Then
				raydir[0] :+ (pix.width-1)*hstep[0]
				raydir[1] :+ (pix.width-1)*hstep[1]
				raydir[2] :+ (pix.width-1)*hstep[2]
				
				raydir[0] :- vstep[0]
				raydir[1] :- vstep[1]
				raydir[2] :- vstep[2]
			EndIf
			For Local x:Int = 0 Until pix.width
				If x > 0 Then
					raydir[0] :- hstep[0] 
					raydir[1] :- hstep[1]
					raydir[2] :- hstep[2]
				EndIf
				
				wld.Ray(pos, raydir, vis, r, g, b)
				
				pixptr[0] = r
				pixptr[1] = g
				pixptr[2] = b
				pixptr[3] = 255
				
				pixptr :+ 4
			Next
		Next
	EndFunction
	
	
	'Pixmap zeichnen
	Function Draw(x:Int, y:Int, s:Int)
		glDisable(GL_TEXTURE_2D)
		glPixelZoom(s,-s)
		glRasterPos2i(0, 0)
		glBitmap(0, 0, 0, 0, x,-y, Null)
		glPixelStorei(GL_UNPACK_ROW_LENGTH, pix.pitch Shr 2)
		glDrawPixels(pix.width, pix.height, GL_RGBA, GL_UNSIGNED_BYTE, pix.pixels)
		glEnable(GL_TEXTURE_2D)
	EndFunction
	
	
	'Bewegt die Kamera
	Function Move(dir:Float[], speed:Float)
		pos[0] :+ dir[0]*speed
		pos[1] :+ dir[1]*speed
		pos[2] :+ dir[2]*speed
	EndFunction
	
	
	'Rotiert die Kamera
	Function Rotate(ang:Float[])
		Local sin0:Float = Sin(ang[0])
		Local cos0:Float = Cos(ang[0])
		Local sin1:Float = Sin(ang[1])
		Local cos1:Float = Cos(ang[1])
		Local tmp:Float[3]
		
		dir = [0.0, 0.0, 1.0]
		upv = [0.0, 1.0, 0.0]
		
		'X-Achse
		tmp[1] = dir[1]*cos0-dir[2]*sin0
		tmp[2] = dir[1]*sin0+dir[2]*cos0
		dir[1] = tmp[1]; dir[2] = tmp[2]
		
		tmp[1] = upv[1]*cos0-upv[2]*sin0
		tmp[2] = upv[1]*sin0+upv[2]*cos0
		upv[1] = tmp[1]; upv[2] = tmp[2]
		'Y-Achse
		tmp[0] = dir[0]*cos1+dir[2]*sin1
		tmp[2] = dir[2]*cos1-dir[0]*sin1
		dir[0] = tmp[0]; dir[2] = tmp[2]
		
		tmp[0] = upv[0]*cos1+upv[2]*sin1
		tmp[2] = upv[2]*cos1-upv[0]*sin1
		upv[0] = tmp[0]; upv[2] = tmp[2]
	EndFunction	
EndType



Type wld
	Global oct:Octree
	Global size:Int 
	Global voxcnt:Int
	
	Global lod:Int
	Global maxlod:Int 
	
	Global rfog:Byte  = 0
	Global gfog:Byte  = 0
	Global bfog:Byte  = 0
	
	Global litx:Float = 0.5844
	Global lity:Float = 0.3061
	Global litz:Float = 0.7514
	
	
	Function Init(inlod:Int)
		oct = New octree
		size = 2^inlod
		lod = inlod
		maxlod = inlod
		
		SeedRnd(21341)
		Generator()
	EndFunction
	
	
	Function Generator()
		Local noise:SimplexNoise = New SimplexNoise
		Local scl:Float = 0.05
		
		For Local z:Int = 0 Until size
			For Local y:Int = 0 Until size
				For Local x:Int = 0 Until size
					Local density:Float = noise.Noise_3D(Float(x)*scl, Float(y)*scl, Float(z)*scl)
					If density > -0.2 And density < 0.2 Then
						oct.Write(x, y, z, size)
						voxcnt :+ 1
					EndIf
				Next
			Next
		Next
	EndFunction
	
	
	Function Quader(x0:Int, y0:Int, z0:Int, x1:Int, y1:Int, z1:Int)
		For Local y:Int = 0 To y1
			For Local x:Int = 0 To x1
				oct.Write(x0+x, y0+y, z0, size)
				oct.Write(x0+x, y0+y, z0+z1, size)
				voxcnt :+ 2
			Next
		Next
		For Local z:Int = 0 To z1
			For Local x:Int = 0 To x1
				oct.Write(x0+x, y0, z0+z, size)
				oct.Write(x0+x, y0+y1, z0+z, size)	
				voxcnt :+ 2
			Next
		Next
		For Local z:Int = 0 To z1
			For Local y:Int = 0 To y1
				oct.Write(x0, y0+y, z0+z, size)
				oct.Write(x0+x1, y0+y, z0+z, size)
				voxcnt :+ 2
			Next
		Next
	EndFunction
	
	
	Function Circle(x0:Int, y0:Int, z0:Int, r:Int)
		Local x1:Int = r
		Local y1:Int = 0
		Local er:Int = r
		Local dx:Int
		Local dy:Int
		
		oct.Write(x0+r, y0, z0, size)
		oct.Write(x0-r, y0, z0, size)
		oct.Write(x0, y0+r, z0, size)
		oct.Write(x0, y0-r, z0, size)
		voxcnt :+ 4
		
		While y1 < x1
			dy = y1*2+1
			er :- dy 
			y1 :+ 1
			If er < 0 Then
				dx = 1-x1*2
				er :- dx
				x1 :- 1
			EndIf
			oct.Write(x0+x1, y0+y1, z0, size)
			oct.Write(x0+x1, y0-y1, z0, size)
			oct.Write(x0-x1, y0+y1, z0, size)
			oct.Write(x0-x1, y0-y1, z0, size)
			oct.Write(x0+y1, y0+x1, z0, size)
			oct.Write(x0+y1, y0-x1, z0, size)
			oct.Write(x0-y1, y0+x1, z0, size)
			oct.Write(x0-y1, y0-x1, z0, size)
			voxcnt :+ 8
		Wend
	EndFunction
	
	
	Function Ray(pos:Float[], dir:Float[], vis:Int, r:Byte Var, g:Byte Var, b:Byte Var)
		Local x:Double = pos[0]
		Local y:Double = pos[1]
		Local z:Double = pos[2]
		
		Local dx:Float = dir[0]
		Local dy:Float = dir[1]
		Local dz:Float = dir[2]
		
		Local tmx:Double
		Local tmy:Double
		Local tmz:Double
		
		Local ix:Int
		Local iy:Int
		Local iz:Int
		
		Local dist:Float
		Local face:Byte
		Local bd:Int[6]
		
		
		Repeat
			If dist >= vis Then
				r = rfog
				g = gfog
				b = bfog 
				Exit
			EndIf
			
			If x >= size Or x <= 0 Or y >= size Or y <= 0 Or z >= size Or z <= 0 Then
				r = 127+(rfog-127)*dist/vis
				g = 127+(gfog-127)*dist/vis
				b = 127+(bfog-127)*dist/vis
				Exit
			EndIf
			
			If dx < 0 Then
				ix = x-0.000000001
			Else
				ix = x
			EndIf
			If dy < 0 Then
				iy = y-0.000000001
			Else
				iy = y
			EndIf
			If dz < 0 Then
				iz = z-0.000000001
			Else
				iz = z
			EndIf
			
			bd = New Int[6]
			If oct.GetState(ix, iy, iz, size, lod, bd) Then 
				Local lit:Int
				Select face
					Case 0  lit = (0.5+litx)*255
					Case 1  lit = (1.5-litx)*255
					Case 2  lit = (0.5+lity)*255
					Case 3  lit = (1.5-lity)*255
					Case 4  lit = (0.5+litz)*255
					Case 5  lit = (1.5-litz)*255
				EndSelect
				If lit > 255 Then lit = 255
				r = lit+(rfog-lit)*dist/vis
				g = lit+(gfog-lit)*dist/vis
				b = lit+(bfog-lit)*dist/vis
				Exit
			EndIf
			
			If dx < 0 Then
				tmx = (x-bd[3])/Abs(dx)
			Else
				tmx = (bd[0]-x)/dx
			EndIf
			If dy < 0 Then
				tmy = (y-bd[4])/Abs(dy)
			Else
				tmy = (bd[1]-y)/dy
			EndIf
			If dz < 0 Then
				tmz = (z-bd[5])/Abs(dz)
			Else
				tmz = (bd[2]-z)/dz
			EndIf
			
			If tmx < tmy And tmx < tmz Then
				x :+ tmx*dx
				y :+ tmx*dy
				z :+ tmx*dz
				
				dist :+ tmx
				
				If dx < 0 Then
					face = 1
				Else
					face = 0
				EndIf
			ElseIf tmy < tmx And tmy < tmz Then
				x :+ tmy*dx
				y :+ tmy*dy
				z :+ tmy*dz
				
				dist :+ tmy
				
				If dy < 0 Then
					face = 3
				Else
					face = 2
				EndIf
			ElseIf tmz < tmx And tmz < tmy Then
				x :+ tmz*dx
				y :+ tmz*dy
				z :+ tmz*dz
				
				dist :+ tmz
				
				If dz < 0 Then
					face = 5
				Else
					face = 4
				EndIf
			EndIf
		Forever
	EndFunction
EndType


Function GetFps:Int()
   Global frame:Int
   Global time:Int
   Global fps:Int
   
	frame = frame + 1
	
	If MilliSecs() >= time Then
		time = MilliSecs()+1000
		fps = frame
		frame = 0
	EndIf 
	Return fps
EndFunction