Type Octree
	Field child:Octree[8]
	Field r:Byte, g:Byte, b:Byte
	
	
	Method Add(x:Int, y:Int, z:Int, r:Byte, g:Byte, b:Byte, size:Int, lvl:Int)
		Write(x, y, z, r, g, b, size)
		For Local i:Int = 1 Until lvl
			TidyColor(x, y, z, size, lvl-i)
		Next 
	EndMethod
	
	
	Method Remove(x:Int, y:Int, z:Int, size:Int, lvl:Int)
		Repeat
			If Not Tidy(x, y, z, size) Then Exit
		Forever
		For Local i:Int = 1 Until lvl
			TidyColor(x, y, z, size, lvl-i)
		Next
	EndMethod
	
	
	Method Write(x:Int, y:Int, z:Int, r:Byte, g:Byte, b:Byte, size:Int)
		Local half:Int = size/2
		If z < half Then
			If y < half Then
				If x < half Then
					WriteOp(0, x     , y     , z     , r, g, b, half)
				Else
					WriteOp(1, x-half, y     , z     , r, g, b, half)
				EndIf
			Else
				If x < half Then
					WriteOp(2, x     , y-half, z     , r, g, b, half)
				Else
					WriteOp(3, x-half, y-half, z     , r, g, b, half)
				EndIf
			EndIf
		Else
			If y < half Then
				If x < half Then
					WriteOp(4, x     , y     , z-half, r, g, b, half)
				Else
					WriteOp(5, x-half, y     , z-half, r, g, b, half)
				EndIf
			Else
				If x < half Then
					WriteOp(6, x     , y-half, z-half, r, g, b, half)
				Else
					WriteOp(7, x-half, y-half, z-half, r, g, b, half)
				EndIf
			EndIf
		EndIf
	EndMethod
	
	
	Method WriteOp(i:Int, x:Int, y:Int, z:Int, r:Byte, g:Byte, b:Byte, size:Int)
		If Not child[i] Then
			child[i] = New Octree
		EndIf
		If size > 1 Then
			child[i].Write(x, y, z, r, g, b, size)
		Else
			child[i].r = r
			child[i].g = g
			child[i].b = b
		EndIf
	EndMethod


	Method Tidy:Byte(x:Int, y:Int, z:Int, size:Int)
		Local half:Int = size/2
		If z < half Then
			If y < half Then
				If x < half Then
					Return TidyOp(0, x     , y     , z     , half)
				Else
					Return TidyOp(1, x-half, y     , z     , half)
				EndIf
			Else
				If x < half Then
					Return TidyOp(2, x     , y-half, z     , half)
				Else
					Return TidyOp(3, x-half, y-half, z     , half)
				EndIf
			EndIf
		Else
			If y < half Then
				If x < half Then
					Return TidyOp(4, x     , y     , z-half, half)
				Else
					Return TidyOp(5, x-half, y     , z-half, half)
				EndIf
			Else
				If x < half Then
					Return TidyOp(6, x     , y-half, z-half, half)
				Else
					Return TidyOp(7, x-half, y-half, z-half, half)
				EndIf
			EndIf
		EndIf
	EndMethod
	
	
	Method TidyOp:Byte(i:Int, x:Int, y:Int, z:Int, size:Int)
		If Not child[i] Then
			Return 0 
		Else
			Local useless:Byte = True
			For Local j:Int = 0 To 7
				If child[i].child[j] <> Null Then useless = False
			Next
			If useless = True Then
				child[i] = Null
				Return 1
			Else
				Return child[i].Tidy(x, y, z, size)
			EndIf
		EndIf
	EndMethod
	
	
	Method TidyColor:Byte(x:Int, y:Int, z:Int, size:Int, lvl:Int)
		Local half:Int = size/2
		If z < half Then
			If y < half Then
				If x < half Then
					Return TidyColorOp(0, x     , y     , z     , half, lvl)
				Else
					Return TidyColorOp(1, x-half, y     , z     , half, lvl)
				EndIf
			Else
				If x < half Then
					Return TidyColorOp(2, x     , y-half, z     , half, lvl)
				Else
					Return TidyColorOp(3, x-half, y-half, z     , half, lvl)
				EndIf
			EndIf
		Else
			If y < half Then
				If x < half Then
					Return TidyColorOp(4, x     , y     , z-half, half, lvl)
				Else
					Return TidyColorOp(5, x-half, y     , z-half, half, lvl)
				EndIf
			Else
				If x < half Then
					Return TidyColorOp(6, x     , y-half, z-half, half, lvl)
				Else
					Return TidyColorOp(7, x-half, y-half, z-half, half, lvl)
				EndIf
			EndIf
		EndIf
	EndMethod
	
	
	Method TidyColorOp:Byte(i:Int, x:Int, y:Int, z:Int, size:Int, lvl:Int)
		If Not child[i] Then
			Return 0
		Else
			If lvl > 1 Then
				Return child[i].TidyColor(x, y, z, size, lvl-1)
			Else
				Local r:Int, g:Int, b:Int, cnt:Int
				For Local j:Int = 0 To 7
					If child[i].child[j] <> Null Then
						r :+ child[i].child[j].r
						g :+ child[i].child[j].g
						b :+ child[i].child[j].b
						cnt :+ 1
					EndIf
				Next
				If cnt > 0 Then
					child[i].r = r/cnt
					child[i].g = g/cnt
					child[i].b = b/cnt
					Return 1
				EndIf
			EndIf
		EndIf
	EndMethod
	
	
	Method GetState:Byte(x:Int, y:Int, z:Int, size:Int, lvl:Int, bd:Int[], r:Byte Var, g:Byte Var, b:Byte Var)
		Local half:Int = size/2
		If z < half Then
			If y < half Then
				If x < half Then
					If Not child[0] Then 
						bd[0] :+ half
						bd[1] :+ half
						bd[2] :+ half
						Return 0
					Else
						If lvl > 1 Then 
							Return child[0].GetState(x, y, z, half, lvl-1, bd, r, g, b)
						Else
							r = child[0].r
							g = child[0].g
							b = child[0].b
							Return 1
						End If
					End If
				Else
					If Not child[1] Then 
						bd[0] :+ size
						bd[1] :+ half
						bd[2] :+ half
						bd[3] :+ half
						Return 0
					Else
						If lvl > 1 Then 
							bd[0] :+ half
							bd[3] :+ half
							Return child[1].GetState(x-half, y, z, half, lvl-1, bd, r, g, b)
						Else
							r = child[1].r
							g = child[1].g
							b = child[1].b
							Return 1
						End If
					End If
				End If
			Else
				If x < half Then
					If Not child[2] Then 
						bd[0] :+ half
						bd[1] :+ size
						bd[2] :+ half
						bd[4] :+ half
						Return 0
					Else
						If lvl > 1 Then 
							bd[1] :+ half
							bd[4] :+ half
							Return child[2].GetState(x, y-half, z, half, lvl-1, bd, r, g, b)
						Else
							r = child[2].r
							g = child[2].g
							b = child[2].b
							Return 1
						End If
					End If
				Else
					If Not child[3] Then 
						bd[0] :+ size
						bd[1] :+ size
						bd[2] :+ half
						bd[3] :+ half
						bd[4] :+ half
						Return 0
					Else
						If lvl > 1 Then 
							bd[0] :+ half
							bd[1] :+ half
							bd[3] :+ half
							bd[4] :+ half
							Return child[3].GetState(x-half, y-half, z, half, lvl-1, bd, r, g, b)
						Else
							r = child[3].r
							g = child[3].g
							b = child[3].b
							Return 1
						End If
					End If
				End If
			End If
		Else
			If y < half Then
				If x < half Then
					If Not child[4] Then 
						bd[0] :+ half
						bd[1] :+ half 
						bd[2] :+ size
						bd[5] :+ half
						Return 0
					Else
						If lvl > 1 Then 
							bd[2] :+ half
							bd[5] :+ half
							Return child[4].GetState(x, y, z-half, half, lvl-1, bd, r, g, b)
						Else
							r = child[4].r
							g = child[4].g
							b = child[4].b
							Return 1
						End If
					End If
				Else
					If Not child[5] Then 
						bd[0] :+ size
						bd[1] :+ half
						bd[2] :+ size
						bd[3] :+ half
						bd[5] :+ half
						Return 0
					Else
						If lvl > 1 Then
							bd[0] :+ half
							bd[2] :+ half
							bd[3] :+ half
							bd[5] :+ half 
							Return child[5].GetState(x-half, y, z-half, half, lvl-1, bd, r, g, b)
						Else
							r = child[5].r
							g = child[5].g
							b = child[5].b
							Return 1
						End If
					End If
				End If
			Else
				If x < half Then
					If Not child[6] Then 
						bd[0] :+ half
						bd[1] :+ size
						bd[2] :+ size
						bd[4] :+ half
						bd[5] :+ half
						Return 0
					Else
						If lvl > 1 Then 
							bd[1] :+ half
							bd[2] :+ half
							bd[4] :+ half
							bd[5] :+ half
							Return child[6].GetState(x, y-half, z-half, half, lvl-1, bd, r, g, b)
						Else
							r = child[6].r
							g = child[6].g
							b = child[6].b
							Return 1
						End If
					End If
				Else
					If Not child[7] Then 
						bd[0] :+ size
						bd[1] :+ size
						bd[2] :+ size
						bd[3] :+ half
						bd[4] :+ half
						bd[5] :+ half
						Return 0
					Else
						If lvl > 1 Then 
							bd[0] :+ half
							bd[1] :+ half
							bd[2] :+ half
							bd[3] :+ half
							bd[4] :+ half
							bd[5] :+ half
							Return child[7].GetState(x-half, y-half, z-half, half, lvl-1, bd, r, g, b)
						Else
							r = child[7].r
							g = child[7].g
							b = child[7].b
							Return 1
						End If
					End If
				End If
			End If
		End If
	End Method
End Type